from __future__ import absolute_import
import copy
import logging
import os
import plotly
from retrying import retry
import requests
import socket
import sys

try:
    import http.client as http_client
except ImportError:
    # Python 2
    import httplib as http_client

try:
    from urlparse import urlparse
except ImportError:
    from urllib.parse import urlparse


# API requests get their config from the environment.
# If variables aren't there, then they check the
# plotly.tools.get_credentials_file


def credential(key):
    if key in os.environ:
        return os.environ[key]
    elif key.upper() in os.environ:
        return os.environ[key.upper()]
    else:
        return plotly.tools.get_credentials_file()[key.replace('plotly_', '')]


def config(key):
    if key in os.environ:
        value = os.environ[key]
    elif key.upper() in os.environ:
        value = os.environ[key.upper()]
    else:
        value = plotly.tools.get_config_file().get(key)

    # Handle PLOTLY_SSL_VERIFICATION which is True or False but a
    # string in environ
    if value == 'False':
        return False
    elif value == 'True':
        return True
    else:
        return value


HEADERS = {
    'plotly-client-platform': 'dash-auth',
    'content-type': 'application/json'
}


def _modify_request_kwargs(request_kwargs):
    copied_kwargs = copy.deepcopy(request_kwargs)
    if 'headers' in request_kwargs:
        copied_kwargs['headers'].update(HEADERS)
    else:
        copied_kwargs['headers'] = HEADERS

    if 'Authorization' not in copied_kwargs['headers']:
        copied_kwargs['auth'] = (
            credential('plotly_username'),
            credential('plotly_api_key'),)

    if 'DASH_STREAMBED_DIRECT_IP' in os.environ:
        copied_kwargs['verify'] = False
    else:
        copied_kwargs['verify'] = config('plotly_ssl_verification')

    return copied_kwargs


def debug_requests_on(url, **kwargs):
    # Adapted from https://stackoverflow.com/a/24588289/4142536
    http_client.HTTPConnection.debuglevel = logging.DEBUG
    logging.basicConfig()
    logging.getLogger().setLevel(logging.DEBUG)
    requests_log = logging.getLogger("requests.packages.urllib3")
    requests_log.setLevel(logging.DEBUG)
    requests_log.propagate = True
    requests_log.addHandler(logging.StreamHandler(sys.stdout))

    # Log some additional info
    print('Attempting connection at {}'.format(url))
    parsed_url = urlparse(url)
    try:
        print('gettaddrinfo: {}'.format(
            socket.getaddrinfo(parsed_url.netloc, parsed_url.scheme)
        ))
    except Exception as e:
        print('gettaddrinfo failed')
        print(e)


def debug_requests_off():
    http_client.HTTPConnection.debuglevel = 0
    requests_log = logging.getLogger("requests.packages.urllib3")
    requests_log.setLevel(logging.NOTSET)
    requests_log.propagate = False


def _create_method(method_name):
    def request(path, api_key_auth=True, **request_kwargs):
        copied_kwargs = _modify_request_kwargs(request_kwargs)
        if 'DASH_STREAMBED_DIRECT_IP' in os.environ:
            base_url = 'https://{}'.format(config('dash_streambed_direct_ip'))
        else:
            base_url = config('plotly_api_domain')

        request_method = getattr(requests, method_name)

        VALID_4XX_STATUS_CODES = [404, 405]

        def check_request_before_returning(url, **kwargs):
            resp = request_method(url, **kwargs)

            # 404's are the only accepted "error" code
            # as we use this to check if a file exists or not
            if resp.status_code not in VALID_4XX_STATUS_CODES:
                resp.raise_for_status()
            return resp

        request_with_retry = retry(
            wait_random_min=100,
            wait_random_max=1000,
            wait_exponential_max=10000,
            stop_max_delay=30000
        )(check_request_before_returning)

        def retry_request_with_logs(url, **kwargs):
            debug_requests_off()
            try:
                return request_with_retry(url, **kwargs)
            except BaseException:
                # request-level errors include ConnectionError
                print(sys.exc_info())

                # do the request one last time with logs
                debug_requests_on(url)
                resp = request_method(url, **kwargs)
                if resp.status_code not in VALID_4XX_STATUS_CODES:
                    resp.raise_for_status()

                # in the off chance that it succeeded on its last request
                # return the response
                return resp

        return retry_request_with_logs(
            '{}{}'.format(base_url, path),
            **copied_kwargs
        )

    return request


post = _create_method('post')
patch = _create_method('patch')
get = _create_method('get')



"""
dash-enterprise auth

Methods to integrate dash apps with the authentication from the
dash-deployment-server.
"""
from .version import __version__
import datetime as _dt
import os as _os
import platform as _platform
import base64 as _b64
import functools as _ft
import json as _json
import jwt as _jwt
import urllib as _urllib
from typing import Any

import dash as _dash
if hasattr(_dash, "dcc"):
    _dcc = _dash.dcc
else:
    import dash_core_components as _dcc
import flask as _flask

logout_url = _os.getenv('DASH_LOGOUT_URL')
ua_string = 'Plotly/%s (Language=Python/%s; Platform=%s/%s)' % (__version__, _platform.python_version(), _platform.system(), _platform.release())


class UaPyJWKClient(_jwt.PyJWKClient):
    def fetch_data(self) -> Any:
        with _urllib.request.urlopen(_urllib.request.Request(self.uri, headers={'User-Agent': ua_string})) as response:
            return _json.load(response)


def _need_request_context(func):
    @_ft.wraps(func)
    def _wrap(*args, **kwargs):
        if not _flask.has_request_context():
            raise RuntimeError('`{0}` method needs a flask/dash request'
                               ' context to run. Make sure to run '
                               '`{0}` from a callback.'.format(func.__name__))
        return func(*args, **kwargs)
    return _wrap


def create_logout_button(label='Logout'):
    """
    Create a dcc.LogoutButton with the dash-deployment-server logout url set
    in the environment.

    :param label: Text of the logout button.
    :type label: str
    :return:
    """
    if not logout_url:
        raise Exception(
            'DASH_LOGOUT_URL was not set in the environment.'
        )

    return _dcc.LogoutButton(
        logout_url=logout_url,
        label=label,
    )


@_need_request_context
def get_user_data():
    jwks_url = _os.getenv('DASH_JWKS_URL')
    if not jwks_url:
        return _json.loads(_flask.request.headers.get('Plotly-User-Data', "{}"))
    try:
        jwks_client = UaPyJWKClient(jwks_url)

        b64token = _flask.request.cookies.get('kcIdToken')
        token = _b64.b64decode(b64token)
        signing_key = jwks_client.get_signing_key_from_jwt(token)

        return _jwt.decode(
            token,
            signing_key.key,
            algorithms=[signing_key._jwk_data.get('alg', 'RSA256')],
            audience=_os.getenv('DASH_AUD', "dash"),
            options={"verify_exp": True},
        )
    except Exception as e:
        print("JWT decode error: " + repr(e))
    return {}


@_need_request_context
def get_username():
    """
    Get the current user.

    :return: The current user.
    :rtype: str
    """
    data = get_user_data()
    if not _os.getenv('DASH_JWKS_URL'):
        return data.get('username')
    return data.get('preferred_username')


@_need_request_context
def get_kerberos_ticket_cache():
    """
    Get the kerberos ticket for the current logged user.

    :return: The kerberos ticket cache.
    """
    data = get_user_data()

    expiry_str = data['kerberos_ticket_expiry']
    expiry = _dt.datetime.strptime(expiry_str, '%Y-%m-%dT%H:%M:%SZ')
    if expiry < _dt.datetime.utcnow():
        raise Exception('Kerberos ticket has expired.')

    return _b64.b64decode(data['kerberos_ticket_cache'])



import io
from setuptools import setup

main_ns = {}
exec(open("dash_enterprise_auth/version.py").read(), main_ns)  # pylint: disable=exec-used

setup(
    name="dash-enterprise-auth",
    version=main_ns["__version__"],
    description="Authentication integrations for dash apps using dash-deployment-server",
    long_description=io.open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    keywords="dash dash-enterprise dash-auth plotly",
    author="Antoine Roy-Gobeil",
    author_email="antoine@plotly.com",
    packages=[
        "dash_enterprise_auth"
    ],
    install_requires=[
        "dash",
        "requests[security]",
        "retrying",
        "PyJWT",
        "cryptography"
    ],
    python_requires=">=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*",
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7"
    ]
)
